#include "sdk_common.h"

#include "ble_db_discovery.h"
#include "ble_gattc.h"
#include "ble_srv_common.h"
#include "ble_types.h"
#include "relay_client.h"
#define NRF_LOG_MODULE_NAME relay_c
#include "nrf_log.h"
NRF_LOG_MODULE_REGISTER();


/**@brief Function for creating a message for writing to the RELAY ACTION. */
void relay_send_write(ble_relay_c_t *p_ble_relay_c, uint16_t handle, uint8_t *p_string, uint16_t length) {
  uint32_t err_code;

  ble_gattc_write_params_t const write_params =
      {
          .write_op = BLE_GATT_OP_PREP_WRITE_REQ,
          .flags = BLE_GATT_EXEC_WRITE_FLAG_PREPARED_WRITE,
          .handle = handle,
          .offset = 0,
          .len = length,
          .p_value = p_string};

  err_code = sd_ble_gattc_write(p_ble_relay_c->conn_handle, &write_params);
  if (err_code == NRF_SUCCESS) {
    NRF_LOG_INFO("relay_send_write: WRITE_REQ(handle = %d, connhandle = %d, error code = %d)", handle, p_ble_relay_c->conn_handle, err_code);
  } else {
    NRF_LOG_INFO("relay_send_write: WRITE_REQ UNSUCCESSFUL(%d)",  err_code);
  }
}

static uint32_t cccd_configure(uint16_t conn_handle, uint16_t cccd_handle, bool enable) {
  uint8_t buf[BLE_CCCD_VALUE_LEN];

  buf[0] = enable ? BLE_GATT_HVX_NOTIFICATION : 0;
  buf[1] = 0;

  ble_gattc_write_params_t const write_params =
      {
          .write_op = BLE_GATT_OP_WRITE_REQ,
          .flags = BLE_GATT_EXEC_WRITE_FLAG_PREPARED_WRITE,
          .handle = cccd_handle,
          .offset = 0,
          .len = sizeof(buf),
          .p_value = buf};
  NRF_LOG_INFO("cccd_configure: cccd handle = %d", cccd_handle);
  return sd_ble_gattc_write(conn_handle, &write_params);
}

uint32_t ble_relay_c_state_notif_enable(ble_relay_c_t *p_ble_relay_c) {
  VERIFY_PARAM_NOT_NULL(p_ble_relay_c);

  if (p_ble_relay_c->conn_handle == BLE_CONN_HANDLE_INVALID) {
    return NRF_ERROR_INVALID_STATE;
  }
  NRF_LOG_INFO("cccd configure with: connection handle (0x%x), cccd handle(0x%x)", p_ble_relay_c->conn_handle, p_ble_relay_c->peer_relay_service_db.relay_cccd_handle);
  return cccd_configure(p_ble_relay_c->conn_handle, p_ble_relay_c->peer_relay_service_db.relay_cccd_handle, true);
}


/**@brief Function for handling Disconnected event received from the SoftDevice.
 *
 * @details This function check if the disconnect event is happening on the link
 *          associated with the current instance of the module, if so it will set its
 *          conn_handle to invalid.
 *
 * @param[in] p_ble_relay_c Pointer to the Relay Service Client structure.
 * @param[in] p_ble_evt                Pointer to the BLE event received.
 */
static void on_disconnected(ble_relay_c_t *p_ble_relay_c, ble_evt_t const *p_ble_evt) {
  if (p_ble_relay_c->conn_handle == p_ble_evt->evt.gap_evt.conn_handle) {
    p_ble_relay_c->conn_handle = BLE_CONN_HANDLE_INVALID;
    p_ble_relay_c->peer_relay_service_db.relay_action_handle = BLE_GATT_HANDLE_INVALID;
    p_ble_relay_c->peer_relay_service_db.relay_state_handle = BLE_GATT_HANDLE_INVALID;
    p_ble_relay_c->peer_relay_service_db.relay_cccd_handle = BLE_GATT_HANDLE_INVALID;
  }
}

void ble_relay_client_on_db_disc_evt(ble_relay_c_t *p_ble_relay_c, ble_db_discovery_evt_t const *p_evt) {
  // Check if the Relay Service was discovered.
  if (p_evt->evt_type == BLE_DB_DISCOVERY_COMPLETE &&
      p_evt->params.discovered_db.srv_uuid.uuid == BLE_UUID_RELAY_SERVICE_UUID &&
      p_evt->params.discovered_db.srv_uuid.type == p_ble_relay_c->uuid_type) {
    ble_relay_client_evt_t evt;

    evt.evt_type = BLE_RELAY_C_EVT_DISCOVERY_COMPLETE;
    evt.conn_handle = p_evt->conn_handle;

    for (uint32_t i = 0; i < p_evt->params.discovered_db.char_count; i++) {
      const ble_gatt_db_char_t *p_char = &(p_evt->params.discovered_db.charateristics[i]);
      switch (p_char->characteristic.uuid.uuid) {
      case BLE_UUID_RELAY_ACTION_UUID:
        evt.peer_db.relay_action_handle = p_char->characteristic.handle_value;
        NRF_LOG_INFO("ble_relay_client_on_db_disc_evt: action handle = %d", evt.peer_db.relay_action_handle);
        break;
      case BLE_UUID_RELAY_STATE_UUID:
        evt.peer_db.relay_state_handle = p_char->characteristic.handle_value;
        evt.peer_db.relay_cccd_handle = p_char->cccd_handle;
        NRF_LOG_INFO("ble_relay_client_on_db_disc_evt: state handle = %d", evt.peer_db.relay_state_handle);
        NRF_LOG_INFO("ble_relay_client_on_db_disc_evt: cccd handle = %d", evt.peer_db.relay_cccd_handle);
        break;

      default:
        break;
      }
    }

    //If the instance has been assigned prior to db_discovery, assign the db_handles
    if (p_ble_relay_c->conn_handle != BLE_CONN_HANDLE_INVALID) {
      if (p_ble_relay_c->peer_relay_service_db.relay_action_handle == BLE_GATT_HANDLE_INVALID) {
        p_ble_relay_c->peer_relay_service_db = evt.peer_db;
      }
    }

    p_ble_relay_c->evt_handler(p_ble_relay_c, &evt);
  }
}

uint32_t ble_relay_client_init(ble_relay_c_t *p_ble_relay_c, ble_relay_c_init_t *p_ble_relay_c_init) {
  uint32_t err_code;
  ble_uuid_t relay_service_uuid;
  ble_uuid128_t relay_service_base_uuid = BLE_UUID_RELAY_SERVICE_BASE_UUID;

  VERIFY_PARAM_NOT_NULL(p_ble_relay_c);
  VERIFY_PARAM_NOT_NULL(p_ble_relay_c_init);
  VERIFY_PARAM_NOT_NULL(p_ble_relay_c_init->evt_handler);

  p_ble_relay_c->peer_relay_service_db.relay_action_handle = BLE_GATT_HANDLE_INVALID;
  p_ble_relay_c->peer_relay_service_db.relay_state_handle = BLE_GATT_HANDLE_INVALID;
  p_ble_relay_c->conn_handle = BLE_CONN_HANDLE_INVALID;
  p_ble_relay_c->evt_handler = p_ble_relay_c_init->evt_handler;

  err_code = sd_ble_uuid_vs_add(&relay_service_base_uuid, &p_ble_relay_c->uuid_type);
  if (err_code != NRF_SUCCESS) {
    return err_code;
  }
  VERIFY_SUCCESS(err_code);

  relay_service_uuid.type = p_ble_relay_c->uuid_type;
  relay_service_uuid.uuid = BLE_UUID_RELAY_SERVICE_UUID;

  return ble_db_discovery_evt_register(&relay_service_uuid);
}

static void on_hvx(ble_relay_c_t *p_ble_relay_c, ble_evt_t const *p_ble_evt) //(ble_nus_c_t * p_ble_relay_c, ble_evt_t const * p_ble_evt)
{
  // HVX can only occur from client sending.
  if ((p_ble_relay_c->peer_relay_service_db.relay_state_handle != BLE_GATT_HANDLE_INVALID) && (p_ble_evt->evt.gattc_evt.params.hvx.handle == p_ble_relay_c->peer_relay_service_db.relay_state_handle) && (p_ble_relay_c->evt_handler != NULL)) {
    ble_relay_client_evt_t ble_relay_c_evt; //ble_relay_c_evt;

    ble_relay_c_evt.evt_type = BLE_RELAY_C_EVT_STATE_EVT;
    ble_relay_c_evt.p_data = (uint8_t *)p_ble_evt->evt.gattc_evt.params.hvx.data;
    ble_relay_c_evt.data_len = p_ble_evt->evt.gattc_evt.params.hvx.len;
    ble_relay_c_evt.conn_handle = p_ble_evt->evt.gattc_evt.conn_handle;

    p_ble_relay_c->evt_handler(p_ble_relay_c, &ble_relay_c_evt);
    NRF_LOG_DEBUG("Client sending data.");
  }
}

void ble_relay_c_on_ble_evt(ble_evt_t const *p_ble_evt, void *p_context) {
  if ((p_context == NULL) || (p_ble_evt == NULL)) {
    return;
  }
  ble_relay_c_t *p_ble_relay_c = (ble_relay_c_t *)p_context;

  switch (p_ble_evt->header.evt_id) {
  case BLE_GATTC_EVT_WRITE_RSP:
    break;

  case BLE_GAP_EVT_DISCONNECTED:
    on_disconnected(p_ble_relay_c, p_ble_evt);
    break;
  case BLE_GATTC_EVT_HVX:
    on_hvx(p_ble_relay_c, p_ble_evt);
    break;

  default:
    break;
  }
}

uint32_t ble_relay_client_handles_assign(ble_relay_c_t *p_ble_relay_c,
    uint16_t conn_handle,
    const relay_service_db_t *p_peer_handles) {
  VERIFY_PARAM_NOT_NULL(p_ble_relay_c);

  p_ble_relay_c->conn_handle = conn_handle;
  if (p_peer_handles != NULL) {
    p_ble_relay_c->peer_relay_service_db.relay_action_handle = p_peer_handles->relay_action_handle;
    p_ble_relay_c->peer_relay_service_db.relay_state_handle = p_peer_handles->relay_state_handle;
    p_ble_relay_c->peer_relay_service_db.relay_cccd_handle = p_peer_handles->relay_cccd_handle;
  }
  return NRF_SUCCESS;
}